use std::ops::Deref;

use crate::api;
use crate::api::ClientAction;
use crate::api::ClientActions;
use crate::api::JoinResponse;
use crate::api::LoginParam;
use crate::api::ShellResponse;
use crate::herd_core::model::Machine;
use crate::herd_core::mqtt::listen_to_device;
use axum::Json;
use axum::extract::FromRequestParts;
use axum::extract::Path;
use axum::http::StatusCode;
use axum_client_ip::ClientIp;
use axum_extra::TypedHeader;
use axum_extra::headers::Authorization;
use axum_extra::headers::authorization::Bearer;
use based_auth::APIUser;
use based_auth::Sessions;
use based_auth::User;
use owl::get;
use owl::prelude::Model;
use owl::query;
use owl::save;
use serde::Deserialize;
use serde_json::json;
use sheepd::DeviceEntry;
use sheepd::DeviceList;

use super::mqtt::is_within_80_seconds;
use super::mqtt::send_msg;

macro_rules! check_admin {
    ($user:ident) => {
        if !$user.read().is_admin() {
            return (
                StatusCode::UNAUTHORIZED,
                Json(api::Result::Err("Invalid credentials")),
            );
        }
    };
}

pub async fn device_osquery(
    Path(device_id): Path<String>,
    APIUser(user): APIUser,
    Json(payload): Json<api::QueryParam>,
) -> (StatusCode, Json<api::Result<String>>) {
    check_admin!(user);

    let machine: Option<Model<Machine>> = get!(device_id);

    if let Some(machine) = machine {
        let resp = send_msg(
            crate::MQTT.get().unwrap(),
            &machine,
            ClientAction::new(ClientActions::OSQuery(payload.query)),
        )
        .await;
        if let Some(resp) = resp.wait_for(std::time::Duration::from_secs(60)).await {
            let r = match resp.response {
                api::ServerResponses::OSQuery(res) => res,
                _ => unreachable!(),
            };
            (StatusCode::OK, Json(api::Result::OkVal(r)))
        } else {
            (
                StatusCode::BAD_GATEWAY,
                Json(api::Result::Err("Did not receive response from device")),
            )
        }
    } else {
        (StatusCode::NOT_FOUND, Json(api::Result::Err("Not Found")))
    }
}

pub async fn device_shell_cmd(
    Path(device_id): Path<String>,
    APIUser(user): APIUser,
    Json(payload): Json<api::ShellParam>,
) -> (StatusCode, Json<api::Result<ShellResponse>>) {
    check_admin!(user);

    let machine: Option<Model<Machine>> = get!(device_id);

    if let Some(machine) = machine {
        let resp = send_msg(
            crate::MQTT.get().unwrap(),
            &machine,
            ClientAction::new(ClientActions::Shell(payload.cmd, payload.cwd)),
        )
        .await;
        if let Some(resp) = resp.wait_for(std::time::Duration::from_secs(60)).await {
            let r = match resp.response {
                api::ServerResponses::Shell(shell_response) => shell_response,
                _ => unreachable!(),
            };
            (StatusCode::OK, Json(api::Result::OkVal(r)))
        } else {
            (
                StatusCode::BAD_GATEWAY,
                Json(api::Result::Err("Did not receive response from device")),
            )
        }
    } else {
        (StatusCode::NOT_FOUND, Json(api::Result::Err("Not Found")))
    }
}

pub async fn device_get_api(
    Path(device_id): Path<String>,
    APIUser(user): APIUser,
) -> (StatusCode, Json<api::Result<DeviceEntry>>) {
    check_admin!(user);

    let machine: Option<Model<Machine>> = get!(device_id.clone());

    if let Some(machine) = machine {
        let api = machine.read();
        let api = DeviceEntry {
            id: device_id.clone(),
            hostname: api.hostname.clone(),
            online: device_online(&device_id),
        };
        (StatusCode::OK, Json(api::Result::OkVal(api)))
    } else {
        let res = api::Result::<DeviceEntry>::Err("Not Found");
        (StatusCode::NOT_FOUND, Json(res))
    }
}

pub fn device_online(id: &String) -> bool {
    crate::ONLINE
        .get()
        .unwrap()
        .get(id)
        .map(|x| is_within_80_seconds(*x.deref()))
        .unwrap_or(false)
}

pub async fn devices_list(APIUser(user): APIUser) -> (StatusCode, Json<api::Result<DeviceList>>) {
    check_admin!(user);

    let machines: Vec<Model<Machine>> = query!(|_| true);

    let mut ret = vec![];

    for mac in machines {
        let id = mac.read().id.to_string().replace("-", "");
        let online_state = device_online(&id);
        ret.push(DeviceEntry {
            id: id,
            hostname: mac.read().hostname.clone(),
            online: online_state,
        });
    }

    (
        StatusCode::OK,
        Json(api::Result::OkVal(DeviceList { devices: ret })),
    )
}

pub async fn login_user(Json(payload): Json<LoginParam>) -> (StatusCode, Json<serde_json::Value>) {
    log::info!("Login attempt for {}", payload.username);
    let u = User::find(&payload.username).await.unwrap();
    if u.read().verify_pw(&payload.password) {
        let ses = u.read().session().await;
        (
            StatusCode::OK,
            Json(json!(api::Result::OkVal(ses.read().token.as_str()))),
        )
    } else {
        (
            StatusCode::FORBIDDEN,
            Json(json!(api::Result::<api::ShellResponse>::Err("invalid"))),
        )
    }
}

/// Route for joining devices (`/join`)
pub async fn join_device(
    ClientIp(ip): ClientIp,
    Json(payload): Json<api::JoinParams>,
) -> (StatusCode, Json<serde_json::Value>) {
    // TODO : check if exists already

    // TODO : validate join token
    log::info!(
        "New device joined: {} [{}]",
        payload.hostname.trim(),
        payload.machine_id
    );

    let machine = Machine::from_join_param(payload.clone());
    let new_token = machine.token.clone();

    listen_to_device(crate::MQTT.get().unwrap(), &payload.machine_id).await;

    save!(machine);

    let i = crate::IDENTITY.get().unwrap();

    (
        StatusCode::OK,
        Json(json!(JoinResponse {
            token: new_token,
            identity: i.public(),
            mqtt: crate::CONFIG.get().unwrap().mqtt.clone()
        })),
    )
}
